// Custom C header specifically for the VisPy project
// This header includes constant definitions that can be used by VisPy to
// support additional features that may not be strictly GL ES 2.0 compatible,
// but will still work. This allows additional features while not promising or
// leading users to think that all GL3+ features are supported.

// GL3
#define GL_MIN                                              0x8007
#define GL_MAX                                              0x8008
